'use strict';

function send(url) {
  return new Promise(function(resolve, reject){
    var xhr = new XMLHttpRequest();
    function ensureReadiness() {
    if(xhr.readyState < 4)
      return;
    if((xhr.status !== 200) && (xhr.status !== 304))
      reject();
    if(xhr.readyState === 4)
      resolve(xhr.response);
    }
    xhr.onreadystatechange = ensureReadiness;
    xhr.open('GET', url, true);
    xhr.timeout = 5000;
    xhr.send('');
  });
}

class NewFirmwarePoller
{
  constructor(callback) {
    this.getLatest = this.getLatest.bind(this);
    this.families = [
      { models: ['ultralite-mk5l'], latest: 0 },
      { models: ['gen5driver-mac'], latest: 0 },
      { models: ['gen5driver-win'], latest: 0 },
    ];
    var funcs = [];
    for (var i in this.families) {
      var lower = encodeURIComponent(this.families[i].models[0].toLowerCase());
      var url = 'http://firmware.motu.com/api/device/' + lower + '/firmware';
      ((url, family) => {
        funcs.push(()=>send(url, 0));
        funcs.push((r)=>{var j = JSON.parse(r); if (j.length) family.latest = j[0]});
      })(url, this.families[i]);
    }
    funcs.push(()=>callback());
    funcs.reduce((prev, cur) => prev.then(cur), Promise.resolve());
  }
  getLatest(model) {
    for (var i in this.families) {
      if (this.families[i].models.indexOf(model) > -1)
        return this.families[i].latest;
    }
  }
}

class Updater {
  constructor() {
    this.sendMsg = this.sendMsg.bind(this);
    this.receiveMsg = this.receiveMsg.bind(this);

    this.isAvailable = this.isAvailable.bind(this);
    this.openImageFromDisk = this.openImageFromDisk.bind(this);
    this.getImageVersion = this.getImageVersion.bind(this);
    this.bootSamba = this.bootSamba.bind(this);
    this.updateFirmware = this.updateFirmware.bind(this);
    this.updateFPGA = this.updateFPGA.bind(this);
    this.start = this.start.bind(this);
    this.refresh = this.refresh.bind(this);
    this.updateLatestAvailable = this.updateLatestAvailable.bind(this);

    this.services = {};
    this.images = {};
    this.started = false;
    this.modalView = new UpdateModalView();
    
    this.latestPoller = new NewFirmwarePoller(this.updateLatestAvailable);
    this.latestFirmware = "";
    this.latestDriver = "";
    this.driverVersion = "";
      
    this.hasFile = false;
    this.serial = this.versionFW = this.fileVersionFW = this.versionFPGA = this.fileVersionFPGA = "";
    this.updating = false;
  }
  updateLatestAvailable() {
    var refresh = false;
    var latest = this.latestPoller.getLatest('ultralite-mk5l');
    if (latest && latest.version) {
      this.latestFirmware = latest.version;
      refresh = true;
    }
    if (/(Mac)/i.test(navigator.platform)) {
      latest = this.latestPoller.getLatest('gen5driver-mac');
      if (latest && latest.version) {
        this.latestDriver = latest.version;
        refresh = true;
      }
    }
    else {
      latest = this.latestPoller.getLatest('gen5driver-win');
      if (latest && latest.version) {
        this.latestDriver = latest.version;
        refresh = true;
      }
    }
    if (refresh)
      this.refresh();
  }
  updateDriverVersion(version) {
    evSys.dispatchASync('driverVersion', version);
    this.driverVersion = version;
    if (!(/(Mac)/i.test(navigator.platform))) {
      if ((version.match(/\./g) || []).length == 3) {
        var parts = version.split('.');
        this.driverVersion = parts[0] + '.' + parts[1] + '.0+' + String(Number(parts[2]) * 10000 + Number(parts[3]));
      }
    }
    this.refresh();
  }
  refresh() {
    this.modelName = ds.getKVArrayItem('kDevName', 0);
    this.versionFW = ds.getKVArrayItem('kDevVersion', 0).toUpperCase();
    this.versionFPGA = ds.getKVArrayItem('kDevVersion', 1).toUpperCase();

    // Serial number is "ULM5", "UL5E" or "UL5L" + lower 3 bytes, UID is MAC with fffe in center
    this.mac = ds.getKVArrayItem('kDeviceMAC', 0);
    this.serial = (ds.getKVArrayItem('kModelID', 0) + this.mac[3].toString(16) + this.mac[4].toString(16) + this.mac[5].toString(16)).toUpperCase();
    this.uid = 
      this.mac[0].toString(16).padStart(2, '0') + 
      this.mac[1].toString(16).padStart(2, '0') + 
      this.mac[2].toString(16).padStart(2, '0') + "fffe" + 
      this.mac[3].toString(16).padStart(2, '0') + 
      this.mac[4].toString(16).padStart(2, '0') + 
      this.mac[5].toString(16).padStart(2, '0');

    if(this.isAvailable()) {

      // always program device's in samba mode, otherwise they could get stuck
      // note, there may not even be a 
      if(!this.updating && (this.services.samba != undefined) && (this.services.samba.length > 0)) {
        if(this.updating = this.updateFirmware(this.services.samba[0].modelName, this.services.samba[0].uid, this.versionFPGA != this.fileVersionFPGA))
          this.modalView.activate(true);
      }

      if(!this.updating) {
        var v = this.getImageVersion();
        this.fileVersionFW = v.firmwareVersion.toUpperCase();
        this.fileVersionFPGA = v.fpgaVersion.toUpperCase();
        this.hasFile = (this.fileVersionFW != "") && (this.fileVersionFPGA != "");

        if(this.hasFile) {
          // If the firmware version matches the file firmware version, maybe update the fpga.
          // The firmware versions must match so we keep the 2 images the same as the file 
          if((this.versionFW == this.fileVersionFW) && 
             (this.versionFPGA != "") && 
             (this.versionFPGA != this.fileVersionFPGA) && 
             (this.services.usb != undefined) &&
             (this.services.usb.length > 0) &&
             (this.services.usb[0].uid == this.uid)) {
               if(this.updating = this.updateFPGA(this.modelName, this.uid)) {
                this.versionFPGA = ""; 
                this.modalView.activate(true);
              } 
          }
        }
      }
    }
    evSys.dispatchASync('refreshUpdateClients');
  }
  start() {
    if(device.hasHostMsgPipe()) {
      device.sendHostMsgPipeMsg("start");
      evSys.addListener('updateChanged', this.refresh);
      this.started = true;
    }
    ds.addListener('kDeviceMAC', this.refresh);
    ds.addListener('kDevName', this.refresh);
    ds.addListener('kDevVersion', this.refresh);
  }

  isAvailable(){
    return this.started;
  }
  openImageFromDisk() {
    this.sendMsg("open");
  }
  openDriverDownload() {
    this.sendMsg("driver")
  }
  getImageVersion() {
    if(this.images[this.modelName] == undefined) {
      this.sendMsg("version&" + this.modelName);
      this.images[this.modelName] = { "firmwareVersion": "", "fpgaVersion": "" };
    }
    return this.images[this.modelName];
  }
  bootSamba() {
    if(this.hasFile && 
      (this.versionFW != this.fileVersionFW) &&
      (this.images[this.modelName] != undefined)) {
        var n = this.services.usb.length;
        for(var i = 0; i < n; ++i) {
          if(this.services.usb[i].uid == this.uid) {
            this.sendMsg("samba&" + this.uid);
            return;
          }
        }
    }
  }
  updateFirmware(modelName, uid, expectFPGAUpdate) {
    if(this.images[modelName] != undefined) {
      var n = this.services.samba.length;
      for(var i = 0; i < n; ++i) {
        if(this.services.samba[i].uid == uid) {
          this.sendMsg("update&firmware&" + modelName + "&" + uid + "&" + (expectFPGAUpdate ? "expectFPGAUpdate" : "noFPGAUpdate"));
          return true;
        }
      }
    }
    return false;
  }
  updateFPGA(modelName, uid) {
    if(this.images[modelName] != undefined) {
      var n = this.services.usb.length;
      for(var i = 0; i < n; ++i) {
        if(this.services.usb[i].uid == uid) {
          this.sendMsg("update&fpga&" + modelName + "&" + uid);
          return true;
        }
      }
    }
    return false;
  }
  // Sends JSON to updater running in objectiveC/C++ or electron
  sendMsg(msg) {
    // messages:
    // "start" : start search for updatable devices
    // "get" : get the service list
    // "open" : run the open file dialog and verify firmwware image
    // "version&modelName" : get the version string for the firmware for specifed model
    // "update&firmware&modelName&uid" : Firmware update model with UID
    // "update&fpga&modelName&uid" : FPGA update model with UID
    if(this.started)
      device.sendHostMsgPipeMsg(msg);
  }

  // Receives JSON from objectiveC/C++ or electron
  receiveMsg(jsonMsg) {
    // messages:
    // services: {"services":{"usb":[{"uid":"0001f2fffefffffe"}],"samba":[{"uid":"0001f2fffefffffe","modelName":"UltraLite-mk5"}]}}
    // firmware file version { "image" : {"modelName": "UltraLite-mk5", "firmwareVersion": "b812", "fpgaVersion": "9" } }
    // updateStatus std::string msg = "{ "updateStatus": "STATUS", "msg": "MSG" }";
    //   STATUS can be "info", "verifying", "updating", "failed", or "success"
    //   MSG:
    //     "info" = anything
    //     "updateFirmware" = percentage
    //     "verifyingFirmware" = percentage
    //     "updateFPGA" = message
    //     "failed" = why?
    //     "success" = "already up to date" or "update completed succesfully"

    if(jsonMsg.dbgout != undefined) {
      console.log(jsonMsg.dbgout);
    }
    // 
    if(jsonMsg.image != undefined) {
      this.images[jsonMsg.image.modelName] = { "firmwareVersion": jsonMsg.image.firmwareVersion, "fpgaVersion": jsonMsg.image.fpgaVersion };
      evSys.dispatchASync('updateChanged');
    }
    else if(jsonMsg.services != undefined) {
      this.services = jsonMsg.services;
      evSys.dispatchASync('updateChanged');
    }
    else if((jsonMsg.updateStatus != undefined) && this.updating) {
      this.modalView.setStatus(jsonMsg.updateStatus, jsonMsg.msg);
      if((jsonMsg.updateStatus == "success") || (jsonMsg.updateStatus == "failed")) {
        // close the socket connection and reset the datastore
        this.updating = false;
        device.wsClose();
      }
    }
  }
}

class UpdateModalView extends View {
  constructor() {
    super(0,0,kFxWidth,kFxHeight);

    this.setTimer = this.setTimer.bind(this);
    this.onTimer = this.onTimer.bind(this);
    this.activate = this.activate.bind(this)  
    this.fpgaProgress = "";
    this.wantsMouse = true;
    this.setBG(kBlackBG);

    this.modal = new ModalCanvas(false);
    this.modal.mdiv.style.zIndex = 1600;
    this.modal.deactivate();
  }
  activate(showIt) {
    if(!showIt) 
      this.modal.deactivate();
    else {
      this.okToClose = false;
      this.modal.activate(this);
      this.requestRender();
    }
  }
  handleMouseDown(xy) {
    if(this.okToClose)
      this.modal.deactivate();
  }
  setStatus(status, msg) {
    this.status = status;
    this.msg = msg;
    this.requestRender();
    if(this.status == "updateFPGA") 
      this.setTimer(1000);
  }

  setTimer(ms) {
    if(this.timerID != undefined)
      clearTimeout(this.timerID);
    this.timerID = setTimeout(this.onTimer, ms);
  }
  onTimer() {
    this.timerID = undefined;
    if(this.status == "updateFPGA") {
      this.fpgaProgress += ".";
      this.requestRender();
      this.setTimer(1000);
    }
    if(this.status == "waitingToClose") {
      this.modal.deactivate();
    }
  }
  render(c){
    c.fillStyle = "rgb(200,202,204)";
    c.font = "Lighter 30px Helvetica";
    c.textAlign = 'center';
    if(this.status == "verifyFirmware") {
      c.fillText("Comparing device firmware and file versions", this.cx, this.cy-15);
      c.fillText(this.msg + " %", this.cx, this.cy+25);
    } 
    else if(this.status == "updateFirmware") {
      c.fillText("Writing firmware to the device", this.cx, this.cy-15);
      c.fillText(this.msg + " %", this.cx, this.cy+25);
    } 
    else if(this.status == "updateFPGA") {
      c.fillText("Writing FPGA to the device", this.cx, this.cy-15);
      c.fillText(this.fpgaProgress, this.cx, this.cy+25);
    } 
    else if(this.status == "success") {
      c.fillText("Success, restarting device", this.cx, this.cy-15);
      //c.fillText("Click to exit", this.cx, this.cy+25);
      this.fpgaProgress = "";
      clearTimeout(this.timerID);
      this.setTimer(5000);
      this.status = "waitingToClose"
      this.okToClose = true;
    }
    else if(this.status == "failed") {
      c.fillText("Unable to write firmware device", this.cx, this.cy-15);
      c.fillText(this.msg, this.cx, this.cy+25);
      c.fillText("Click to exit", this.cx, this.cy+175);
      this.fpgaProgress = "";
      this.okToClose = true;
    }
    else {
      c.fillText("Configuring device for next update step", this.cx, this.cy-15);
    } 
  }
}


class UpdateView extends View {
  constructor(l, t, m) {
    m = m ? m : 0;
    super(l, t, 370, 180 + m);
    this.margin = m;
    
    // called by the updater
    evSys.addListener('refreshUpdateClients', this.requestRender);

    this.setBG(kEmptyBG);

    this.wantsMouse = true;
    this.chooseRect = new Rect(this.l + this.margin + 160, this.t + 100, 70, 30);
    this.driverRect = new Rect(this.l + this.margin + 160, this.t + 150, 70, 30);
  }

  handleMouseDown(xy) {
    if(updater.isAvailable() && device.isConnected()) {
      if(this.chooseRect.contains(xy.x, xy.y)) {
        updater.openImageFromDisk();
      }
      if(this.driverRect.contains(xy.x, xy.y)) {
        updater.openDriverDownload();
      }
    }
  }

  render(c) {
    if(!device.isConnected())
      return;
    
    var fullFileVer = updater.fileVersionFW + ":" + updater.fileVersionFPGA;

    c.fillStyle = "white";
    c.font = "300 15px Arial";
    c.textAlign = 'left';
    c.fillText("Serial Number:", this.l + this.margin, this.t + 20);
    c.fillText("Firmware Version:", this.l + this.margin, this.t + 70);
    c.fillText("Firmware Updater:", this.l + this.margin, this.t + 120);
    c.fillText("CueMix Download:", this.l + this.margin, this.t + 170);
    c.fillStyle = kToggleDefault;
    c.fillText(updater.serial.toUpperCase(), this.l + this.margin + 160, this.t + 20);
    var updateAvailable = false;
    var verStr = "";
    verStr += (updater.versionFW + ":" + updater.versionFPGA).toUpperCase();
    if(verStr == ":") {
        verStr = "Needs Update";
        updateAvailable = true;
    }
    else if (updater.latestFirmware && (updater.latestFirmware != updater.versionFW))
        updateAvailable = true;
        
    if(updater.hasFile) {
      if(updater.versionFW == updater.fileVersionFW)
        verStr += " up to date";
      else 
        verStr += " updatable";
    }
    c.fillText(verStr, this.l + this.margin  + 160, this.t + 70);

    if(updater.isAvailable()) {
      drawButton(c, this.chooseRect, "Open...", false, "black");
      drawButton(c, this.driverRect, "Open...", false, "black");
      c.fillStyle = "RGB(20,122,255)";
      c.textAlign = 'center';
      if (updateAvailable) {
        c.fillText("Update Available!", this.l + this.margin + 300, this.t + 120);
      }
      if (updater.latestDriver && updater.driverVersion && (updater.latestDriver != updater.driverVersion)) {
        c.fillText("Update Available!", this.l + this.margin + 300, this.t + 170);
      }
    }
  }
}
