'use strict';

////////// helper functions //////////////

function getSampleRateMult(sr) {
   return (sr <= 48000) ? 1 : ((sr <= 96000) ? 2 : 4);
}

// convert float to/from 8.24 int
const k824_1 = 0x01000000;
function To824(v) {
  return Math.round(v * k824_1);
}
function From824(v) {
  return v / k824_1;
}

function From_dB(v) {
  return Math.pow(10, 0.05 * v);
};
const k20OverLog10 = 20 / Math.log(10);
function To_dB(v) {
  return k20OverLog10 * Math.log(v);
};

// convert a float 0-1 to 16bits 0-ffff
function To16(v) {
  return Math.round(k16Bit * v) & k16Bit;
}
// convert a 16bit 0-ffff to a float range 0-1
const k16Bit = 0x0000ffff;
function From16(v) {
  return (v & k16Bit) / k16Bit;
}

function CheckArg(arg, argType) {
    if(!(arg instanceof argType))
      throw 'Bad Arg expected ' + argType.name;
}

function trimFloat(v, nd) {
  if(!isNaN(v))
    v = v.toFixed(nd);
  return v;
}

// updtime
var perf = window.performance;
function perfSeconds() { return perf.now()/1000.0; }

/* View in fullscreen */
function enterFullscreen() {
  var elem = document.documentElement;
  if (elem.requestFullscreen) {
    elem.requestFullscreen();
  } else if (elem.mozRequestFullScreen) { /* Firefox */
    elem.mozRequestFullScreen();
  } else if (elem.webkitRequestFullscreen) { /* Chrome, Safari and Opera */
    elem.webkitRequestFullscreen();
  } else if (elem.msRequestFullscreen) { /* IE/Edge */
    elem.msRequestFullscreen();
  }
}

/* Close fullscreen */
function exitFullscreen() {
  var elem = document.documentElement;
  if (document.exitFullscreen) {
    document.exitFullscreen();
  } else if (document.mozCancelFullScreen) { /* Firefox */
    document.mozCancelFullScreen();
  } else if (document.webkitExitFullscreen) { /* Chrome, Safari and Opera */
    document.webkitExitFullscreen();
  } else if (document.msExitFullscreen) { /* IE/Edge */
    document.msExitFullscreen();
  }
}

function getOutputPairChName(i) {
  const defaults = {
    stereo: {
      0: 'Main 1-2',
      2: 'Line 3-4',
      4: 'Line 5-6',
      6: 'Line 7-8',
      8: 'Line 9-10',
      10: 'Phones 1-2'
    },
    mono: {
      0: 'Main 1',
      1: 'Main 2',
      2: 'Line 3',
      3: 'Line 4',
      4: 'Line 5',
      5: 'Line 6',
      6: 'Line 7',
      7: 'Line 8',
      8: 'Line 9',
      9: 'Line 10',
      10: 'Phones 1',
      11: 'Phones 2'
    }
  };
  if ([0,2,4,6,8,10].includes(i)) {
    var name = ds.getKVArrayItem('koChannelNames', i);
    if (ds.getKVArrayItem('koMixStereo', i)) {
      name = (name ? name : defaults.stereo[i]) + " Mix (L-R)";
    }
    else {
      var other = ds.getKVArrayItem('koChannelNames', i+1);
      other = (other ? other : defaults.mono[i+1]) + " Mix (R)";
      name = (name ? name : defaults.mono[i]) + " Mix (L) - " + other;
    }
    return name;
  } else {
    return "Invalid Index (" + i + ")";
  } 
}

////////// utility classes //////////////
class AnimationFrameDispatcher {
  constructor() {
    this.queue = this.queue.bind(this);
    this.queueMeters = this.queueMeters.bind(this);
    this.dispatch = this.dispatch.bind(this);
    this.metersCB = null;
    this.callbacks = [];
    this.updateTS = 0; 
    window.requestAnimationFrame(this.dispatch);
  }
  queueMeters(cb) {
    this.metersCB = cb;
  }
  queue(cb) {
    if(this.callbacks.indexOf(cb) == -1) {
      this.callbacks.push(cb);
    }
  }
  dispatch(timestamp) {
    timestamp /= 1000.0
    if(this.updateTS <= timestamp) {
      // do meters first so they can be rendered this pass
      if(this.metersCB != null)
        this.metersCB(timestamp);
      
      var c = this.callbacks;
      this.callbacks = [];
      c.forEach(function(cb) { cb(timestamp) });

      this.updateTS = timestamp + kMaxRefreshInterval;
    }
    window.requestAnimationFrame(this.dispatch);
  } 
}


class Events {
  constructor() {
    this.addListener = this.addListener.bind(this);
    this.removeListener = this.removeListener.bind(this);
    this.dispatchSync = this.dispatchSync.bind(this);
    this.dispatchASync = this.dispatchASync.bind(this);
    this.dispatchPending = this.dispatchPending.bind(this);

    this.events = {}; // properties will be the event key holding an array of callbacks for the event
    this.pending = {};
    this.timerQueued = false;
  }
  // FWIW, the callback will be called with the key and the value.
  addListener(ev, l) {
    if(l == undefined)
      throw 'bad callback';

    if(!this.events.hasOwnProperty(ev))
      this.events[ev] = [];
    var o = this.events[ev];
    if(o.indexOf(l) == -1) 
      o.push(l);
  }
  removeListener(ev, l) {
    var o = this.events[ev];
    if(o != undefined) {
      var i = o.indexOf(l);
      if(i > -1) 
        o.splice(i, 1);

      if(o.length == 0)
        delete this.events.ev;
    }
  }
  dispatchSync(ev, v, setByObject) {
    var o = this.events[ev];
    if(o != undefined) {
      var n = o.length;
      for(var i = 0; i < n; i++) 
        o[i](ev, v, setByObject);
    }
  }
  dispatchASync(ev, v, setByObject) {
    this.pending[ev] = [v, setByObject];
    if(!this.timerQueued) {
      this.timerQueued = true;
      setTimeout(this.dispatchPending, 0);
    }
  }
  dispatchPending() {
    var p = this.pending;
    this.timerQueued = false;
    this.pending = {};
    for (let [ev, v] of Object.entries(p)) 
      this.dispatchSync(ev, v[0], v[1]);
  }
}


class Rect {
  constructor(l, t, w, h) {
    this.uRBCXCY = this.uRBCXCY.bind(this);
    this.setR = this.setR.bind(this);
    this.setL = this.setL.bind(this);
    this.setT = this.setT.bind(this);
    this.setLT = this.setLT.bind(this);
    this.offset = this.offset.bind(this);
    this.setW = this.setW.bind(this);
    this.setH = this.setH.bind(this);
    this.setWH = this.setWH.bind(this);
    this.setLTWH = this.setLTWH.bind(this);
    this.contains = this.contains.bind(this);
    this.intersects = this.intersects.bind(this);

    this.l = (l == undefined) ? 0 : l;
    this.t = (t == undefined) ? 0 : t;
    this.w = (w == undefined) ? 0 : w;
    this.h = (h == undefined) ? 0 : h;
    this.r = this.l + this.w;
    this.b = this.t + this.h;
    this.cx = this.l + this.w/2;
    this.cy = this.t + this.h/2;
  }
  uRBCXCY() {
    this.r = this.l + this.w;
    this.b = this.t + this.h;
    this.cx = this.l + this.w/2;
    this.cy = this.t + this.h/2;
  }
  copyR() {
    return new Rect(this.l, this.t, this.w, this.h);
  }
  setR(r) {
    this.setLTWH(r.l, r.t, r.w, r.h);
  }
  setL(l)
  {
    this.l = l;
    this.uRBCXCY();
  }
  setT(t)
  {
    this.t = t;
    this.uRBCXCY();
  }
  setLT(l,t) {
    this.l = l;
    this.t = t;
    this.uRBCXCY();
  }
  offset(dx,dy) {
    this.l = this.l + dx;
    this.t = this.t + dy;
    this.uRBCXCY();
  }
  setW(w) {
    this.w = w;
    this.uRBCXCY();
  }
  setH(h) {
    this.h = h;
    this.uRBCXCY();
  }
  setWH(w,h) {
    this.w = w;
    this.h = h;
    this.uRBCXCY();
  }
  setLTWH(l,t,w,h) {
    this.l = l;
    this.t = t;
    this.w = w;
    this.h = h;
    this.uRBCXCY();
  }
  contains(x, y) {
    return (this.l <= x) && (this.r > x) && (this.t <= y) && (this.b > y);
  }
  containsR(r) {
    return (this.l <= r.l) && (this.r >= r.r) && (this.t <= r.t) && (this.b >= r.b);
  }
  intersects(r) {
    return ! ((this.r < r.l) || (this.l > r.r) || (this.b < r.t) || (this.t > r.b));
  }
}



class List {
  constructor() { 
    this.items = []; 
    this.clear = this.clear.bind(this);
    this.add = this.add.bind(this);
    this.get = this.get.bind(this);
    this.length = this.length.bind(this);
  }  
  clear() { this.items = []; }
  add(v) { if(this.items.indexOf(v) == -1) { this.items.push(v); } }
  remove(v) { 
    var i = this.items.indexOf(v);
    if(i != -1) 
      this.items.splice(i,1);  
  }
  get(i) { return this.items[i]; }
  length() { return this.items.length; }
}



class KeyDownFilter extends List {
  constructor() {
    super();
    this.keyDown = this.keyDown.bind(this);
    document.addEventListener('keydown', this.keyDown);
  }

  keyDown(e) {
    e = e || window.event;

    var keyCode = undefined;
    if (e.keyCode == 13) // escape
        keyCode = "enter";
    if (e.keyCode == 27) // escape
        keyCode = "escape";
    if (e.keyCode == 37) // larrow
        keyCode = "larrow";
    if (e.keyCode == 39) // rarrow
        keyCode = "rarrow";

    if (keyCode != undefined) {
      // go through the list backwards to find the last handler
      var n = this.length();
      for(var i = n-1; i >= 0; i--) {
        if(this.get(i).filterKey(keyCode))
          return;
      }
    }
  }
}

function getExp(cur, exp) {
  if(exp == "toggle")
    return (cur == "on") ? "off" : "on";
  if(exp == "on")
    return "on"
  return "off";
}

class Storage {
  constructor() {
  }
  getItem(key) {
    return localStorage.getItem(key);
  }
  setItem(key, value) {
    localStorage.setItem(key, value);
  }
}